package sun.net.httpserver;

import java.io.OutputStream;
import java.io.InputStream;
import com.sun.net.httpserver.HttpsParameters;
import javax.net.ssl.SSLEngineResult;
import java.nio.ByteBuffer;
import javax.net.ssl.SSLException;
import javax.net.ssl.SSLParameters;
import com.sun.net.httpserver.HttpsConfigurator;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.Lock;
import javax.net.ssl.SSLEngine;
import java.nio.channels.SocketChannel;
import javax.net.ssl.SSLContext;

class SSLStreams
{
    SSLContext sslctx;
    SocketChannel chan;
    TimeSource time;
    ServerImpl server;
    SSLEngine engine;
    SSLStreams.EngineWrapper wrapper;
    SSLStreams.OutputStream os;
    SSLStreams.InputStream is;
    Lock handshaking;
    int app_buf_size;
    int packet_buf_size;
    
    SSLStreams(ServerImpl serverImpl, SSLContext sslctx, SocketChannel chan) throws IOException {
        this.handshaking = new ReentrantLock();
        this.server = serverImpl;
        this.time = serverImpl;
        this.sslctx = sslctx;
        this.chan = chan;
        InetSocketAddress inetSocketAddress = (InetSocketAddress)chan.socket().getRemoteSocketAddress();
        (this.engine = sslctx.createSSLEngine(inetSocketAddress.getHostName(), inetSocketAddress.getPort())).setUseClientMode(false);
        this.configureEngine(serverImpl.getHttpsConfigurator(), inetSocketAddress);
        this.wrapper = new SSLStreams.EngineWrapper(chan, this.engine);
    }
    
    private void configureEngine(HttpsConfigurator httpsConfigurator, InetSocketAddress inetSocketAddress) {
        if (httpsConfigurator != null) {
            SSLStreams.Parameters params = new SSLStreams.Parameters(httpsConfigurator, inetSocketAddress);
            httpsConfigurator.configure(params);
            SSLParameters sslParameters = params.getSSLParameters();
            if (sslParameters != null) {
                this.engine.setSSLParameters(sslParameters);
            }
            else {
                if (params.getCipherSuites() != null) {
                    try {
                        this.engine.setEnabledCipherSuites(params.getCipherSuites());
                    }
                    catch (IllegalArgumentException ex) {}
                }
                this.engine.setNeedClientAuth(params.getNeedClientAuth());
                this.engine.setWantClientAuth(params.getWantClientAuth());
                if (params.getProtocols() != null) {
                    try {
                        this.engine.setEnabledProtocols(params.getProtocols());
                    }
                    catch (IllegalArgumentException ex2) {}
                }
            }
        }
    }
    
    void close() throws IOException {
        this.wrapper.close();
    }
    
    SSLStreams.InputStream getInputStream() throws IOException {
        if (this.is == null) {
            this.is = new SSLStreams.InputStream();
        }
        return this.is;
    }
    
    SSLStreams.OutputStream getOutputStream() throws IOException {
        if (this.os == null) {
            this.os = new SSLStreams.OutputStream();
        }
        return this.os;
    }
    
    SSLEngine getSSLEngine() {
        return this.engine;
    }
    
    void beginHandshake() throws SSLException {
        this.engine.beginHandshake();
    }
    
    private ByteBuffer allocate(SSLStreams.BufType bufType) {
        return this.allocate(bufType, -1);
    }
    
    private ByteBuffer allocate(SSLStreams.BufType bufType, int n) {
        assert this.engine != null;
        synchronized (this) {
            int capacity;
            if (bufType == SSLStreams.BufType.PACKET) {
                if (this.packet_buf_size == 0) {
                    this.packet_buf_size = this.engine.getSession().getPacketBufferSize();
                }
                if (n > this.packet_buf_size) {
                    this.packet_buf_size = n;
                }
                capacity = this.packet_buf_size;
            }
            else {
                if (this.app_buf_size == 0) {
                    this.app_buf_size = this.engine.getSession().getApplicationBufferSize();
                }
                if (n > this.app_buf_size) {
                    this.app_buf_size = n;
                }
                capacity = this.app_buf_size;
            }
            return ByteBuffer.allocate(capacity);
        }
    }
    
    private ByteBuffer realloc(ByteBuffer src, boolean b, SSLStreams.BufType bufType) {
        synchronized (this) {
            ByteBuffer allocate = this.allocate(bufType, 2 * src.capacity());
            if (b) {
                src.flip();
            }
            allocate.put(src);
            src = allocate;
        }
        return src;
    }
    
    public SSLStreams.WrapperResult sendData(ByteBuffer byteBuffer) throws IOException {
        SSLStreams.WrapperResult wrapAndSend = null;
        while (byteBuffer.remaining() > 0) {
            wrapAndSend = this.wrapper.wrapAndSend(byteBuffer);
            if (wrapAndSend.result.getStatus() == SSLEngineResult.Status.CLOSED) {
                this.doClosure();
                return wrapAndSend;
            }
            SSLEngineResult.HandshakeStatus handshakeStatus = wrapAndSend.result.getHandshakeStatus();
            if (handshakeStatus != SSLEngineResult.HandshakeStatus.FINISHED && handshakeStatus != SSLEngineResult.HandshakeStatus.NOT_HANDSHAKING) {
                this.doHandshake(handshakeStatus);
            }
            else {
                continue;
            }
        }
        return wrapAndSend;
    }
    
    public SSLStreams.WrapperResult recvData(ByteBuffer byteBuffer) throws IOException {
        SSLStreams.WrapperResult recvAndUnwrap = null;
        assert byteBuffer.position() == 0;
        while (byteBuffer.position() == 0) {
            recvAndUnwrap = this.wrapper.recvAndUnwrap(byteBuffer);
            byteBuffer = ((recvAndUnwrap.buf != byteBuffer) ? recvAndUnwrap.buf : byteBuffer);
            if (recvAndUnwrap.result.getStatus() == SSLEngineResult.Status.CLOSED) {
                this.doClosure();
                return recvAndUnwrap;
            }
            SSLEngineResult.HandshakeStatus handshakeStatus = recvAndUnwrap.result.getHandshakeStatus();
            if (handshakeStatus != SSLEngineResult.HandshakeStatus.FINISHED && handshakeStatus != SSLEngineResult.HandshakeStatus.NOT_HANDSHAKING) {
                this.doHandshake(handshakeStatus);
            }
            else {
                continue;
            }
        }
        byteBuffer.flip();
        return recvAndUnwrap;
    }
    
    void doClosure() throws IOException {
        try {
            this.handshaking.lock();
            ByteBuffer allocate = this.allocate(SSLStreams.BufType.APPLICATION);
            do {
                allocate.clear();
                allocate.flip();
            } while (this.wrapper.wrapAndSendX(allocate, true).result.getStatus() != SSLEngineResult.Status.CLOSED);
        }
        finally {
            this.handshaking.unlock();
        }
    }
    
    void doHandshake(SSLEngineResult.HandshakeStatus handshakeStatus) throws IOException {
        try {
            this.handshaking.lock();
            ByteBuffer byteBuffer = this.allocate(SSLStreams.BufType.APPLICATION);
            while (handshakeStatus != SSLEngineResult.HandshakeStatus.FINISHED && handshakeStatus != SSLEngineResult.HandshakeStatus.NOT_HANDSHAKING) {
                SSLStreams.WrapperResult wrapperResult = null;
                switch (handshakeStatus) {
                    case NEED_TASK: {
                        Runnable delegatedTask;
                        while ((delegatedTask = this.engine.getDelegatedTask()) != null) {
                            delegatedTask.run();
                        }
                    }
                    case NEED_WRAP: {
                        byteBuffer.clear();
                        byteBuffer.flip();
                        wrapperResult = this.wrapper.wrapAndSend(byteBuffer);
                        break;
                    }
                    case NEED_UNWRAP: {
                        byteBuffer.clear();
                        wrapperResult = this.wrapper.recvAndUnwrap(byteBuffer);
                        if (wrapperResult.buf != byteBuffer) {
                            byteBuffer = wrapperResult.buf;
                        }
                        assert byteBuffer.position() == 0;
                        break;
                    }
                }
                handshakeStatus = wrapperResult.result.getHandshakeStatus();
            }
        }
        finally {
            this.handshaking.unlock();
        }
    }
    
    static ByteBuffer access$000(SSLStreams sslStreams, SSLStreams.BufType bufType) {
        return sslStreams.allocate(bufType);
    }
    
    static ByteBuffer access$100(SSLStreams sslStreams, ByteBuffer byteBuffer, boolean b, SSLStreams.BufType bufType) {
        return sslStreams.realloc(byteBuffer, b, bufType);
    }
    
    class Parameters extends HttpsParameters
    {
        InetSocketAddress addr;
        HttpsConfigurator cfg;
        SSLParameters params;
        
        Parameters(HttpsConfigurator cfg, InetSocketAddress addr) {
            this.addr = addr;
            this.cfg = cfg;
        }
        
        @Override
        public InetSocketAddress getClientAddress() {
            return this.addr;
        }
        
        @Override
        public HttpsConfigurator getHttpsConfigurator() {
            return this.cfg;
        }
        
        @Override
        public void setSSLParameters(SSLParameters params) {
            this.params = params;
        }
        
        SSLParameters getSSLParameters() {
            return this.params;
        }
    }
    
    class WrapperResult
    {
        SSLEngineResult result;
        ByteBuffer buf;
    }
    
    enum BufType
    {
        PACKET, 
        APPLICATION;
    }
    
    class EngineWrapper
    {
        SocketChannel chan;
        SSLEngine engine;
        Object wrapLock;
        Object unwrapLock;
        ByteBuffer unwrap_src;
        ByteBuffer wrap_dst;
        boolean closed;
        int u_remaining;
        
        EngineWrapper(SocketChannel chan, SSLEngine engine) throws IOException {
            this.closed = false;
            this.chan = chan;
            this.engine = engine;
            this.wrapLock = new Object();
            this.unwrapLock = new Object();
            this.unwrap_src = SSLStreams.access$000(SSLStreams.this, SSLStreams.BufType.PACKET);
            this.wrap_dst = SSLStreams.access$000(SSLStreams.this, SSLStreams.BufType.PACKET);
        }
        
        void close() throws IOException {
        }
        
        SSLStreams.WrapperResult wrapAndSend(ByteBuffer byteBuffer) throws IOException {
            return this.wrapAndSendX(byteBuffer, false);
        }
        
        SSLStreams.WrapperResult wrapAndSendX(ByteBuffer src, boolean b) throws IOException {
            if (this.closed && !b) {
                throw new IOException("Engine is closed");
            }
            SSLStreams.WrapperResult wrapperResult = new SSLStreams.WrapperResult();
            synchronized (this.wrapLock) {
                this.wrap_dst.clear();
                SSLEngineResult.Status status;
                do {
                    wrapperResult.result = this.engine.wrap(src, this.wrap_dst);
                    status = wrapperResult.result.getStatus();
                    if (status == SSLEngineResult.Status.BUFFER_OVERFLOW) {
                        this.wrap_dst = SSLStreams.access$100(SSLStreams.this, this.wrap_dst, true, SSLStreams.BufType.PACKET);
                    }
                    else {
                        continue;
                    }
                } while (status == SSLEngineResult.Status.BUFFER_OVERFLOW);
                if (status == SSLEngineResult.Status.CLOSED && !b) {
                    this.closed = true;
                    return wrapperResult;
                }
                if (wrapperResult.result.bytesProduced() > 0) {
                    this.wrap_dst.flip();
                    int i = this.wrap_dst.remaining();
                    assert i == wrapperResult.result.bytesProduced();
                    while (i > 0) {
                        i -= this.chan.write(this.wrap_dst);
                    }
                }
            }
            return wrapperResult;
        }
        
        SSLStreams.WrapperResult recvAndUnwrap(ByteBuffer buf) throws IOException {
            SSLEngineResult.Status ok = SSLEngineResult.Status.OK;
            SSLStreams.WrapperResult wrapperResult = new SSLStreams.WrapperResult();
            wrapperResult.buf = buf;
            if (this.closed) {
                throw new IOException("Engine is closed");
            }
            int n;
            if (this.u_remaining > 0) {
                this.unwrap_src.compact();
                this.unwrap_src.flip();
                n = 0;
            }
            else {
                this.unwrap_src.clear();
                n = 1;
            }
            synchronized (this.unwrapLock) {
                SSLEngineResult.Status status;
                do {
                    if (n != 0) {
                        int i;
                        do {
                            i = this.chan.read(this.unwrap_src);
                        } while (i == 0);
                        if (i == -1) {
                            throw new IOException("connection closed for reading");
                        }
                        this.unwrap_src.flip();
                    }
                    wrapperResult.result = this.engine.unwrap(this.unwrap_src, wrapperResult.buf);
                    status = wrapperResult.result.getStatus();
                    if (status == SSLEngineResult.Status.BUFFER_UNDERFLOW) {
                        if (this.unwrap_src.limit() == this.unwrap_src.capacity()) {
                            this.unwrap_src = SSLStreams.access$100(SSLStreams.this, this.unwrap_src, false, SSLStreams.BufType.PACKET);
                        }
                        else {
                            this.unwrap_src.position(this.unwrap_src.limit());
                            this.unwrap_src.limit(this.unwrap_src.capacity());
                        }
                        n = 1;
                    }
                    else if (status == SSLEngineResult.Status.BUFFER_OVERFLOW) {
                        wrapperResult.buf = SSLStreams.access$100(SSLStreams.this, wrapperResult.buf, true, SSLStreams.BufType.APPLICATION);
                        n = 0;
                    }
                    else {
                        if (status == SSLEngineResult.Status.CLOSED) {
                            this.closed = true;
                            wrapperResult.buf.flip();
                            return wrapperResult;
                        }
                        continue;
                    }
                } while (status != SSLEngineResult.Status.OK);
            }
            this.u_remaining = this.unwrap_src.remaining();
            return wrapperResult;
        }
    }
    
    class InputStream extends java.io.InputStream
    {
        ByteBuffer bbuf;
        boolean closed;
        boolean eof;
        boolean needData;
        byte[] single;
        
        InputStream() {
            this.closed = false;
            this.eof = false;
            this.needData = true;
            this.single = new byte[1];
            this.bbuf = SSLStreams.access$000(SSLStreams.this, SSLStreams.BufType.APPLICATION);
        }
        
        @Override
        public int read(byte[] dst, int offset, int length) throws IOException {
            if (this.closed) {
                throw new IOException("SSL stream is closed");
            }
            if (this.eof) {
                return 0;
            }
            int n = 0;
            if (!this.needData) {
                n = this.bbuf.remaining();
                this.needData = (n == 0);
            }
            if (this.needData) {
                this.bbuf.clear();
                SSLStreams.WrapperResult recvData = SSLStreams.this.recvData(this.bbuf);
                this.bbuf = ((recvData.buf == this.bbuf) ? this.bbuf : recvData.buf);
                if ((n = this.bbuf.remaining()) == 0) {
                    this.eof = true;
                    return 0;
                }
                this.needData = false;
            }
            if (length > n) {
                length = n;
            }
            this.bbuf.get(dst, offset, length);
            return length;
        }
        
        @Override
        public int available() throws IOException {
            return this.bbuf.remaining();
        }
        
        @Override
        public boolean markSupported() {
            return false;
        }
        
        @Override
        public void reset() throws IOException {
            throw new IOException("mark/reset not supported");
        }
        
        @Override
        public long skip(long n) throws IOException {
            int i = (int)n;
            if (this.closed) {
                throw new IOException("SSL stream is closed");
            }
            if (this.eof) {
                return 0L;
            }
            while (i > 0) {
                if (this.bbuf.remaining() >= i) {
                    this.bbuf.position(this.bbuf.position() + i);
                    return (long)i;
                }
                i -= this.bbuf.remaining();
                this.bbuf.clear();
                SSLStreams.WrapperResult recvData = SSLStreams.this.recvData(this.bbuf);
                this.bbuf = ((recvData.buf == this.bbuf) ? this.bbuf : recvData.buf);
            }
            return (long)i;
        }
        
        @Override
        public void close() throws IOException {
            this.eof = true;
            SSLStreams.this.engine.closeInbound();
        }
        
        @Override
        public int read(byte[] array) throws IOException {
            return this.read(array, 0, array.length);
        }
        
        @Override
        public int read() throws IOException {
            if (this.read(this.single, 0, 1) == 0) {
                return -1;
            }
            return this.single[0] & 0xFF;
        }
    }
    
    class OutputStream extends java.io.OutputStream
    {
        ByteBuffer buf;
        boolean closed;
        byte[] single;
        
        OutputStream() {
            this.closed = false;
            this.single = new byte[1];
            this.buf = SSLStreams.access$000(SSLStreams.this, SSLStreams.BufType.APPLICATION);
        }
        
        @Override
        public void write(int n) throws IOException {
            this.single[0] = (byte)n;
            this.write(this.single, 0, 1);
        }
        
        @Override
        public void write(byte[] array) throws IOException {
            this.write(array, 0, array.length);
        }
        
        @Override
        public void write(byte[] src, int offset, int i) throws IOException {
            if (this.closed) {
                throw new IOException("output stream is closed");
            }
            while (i > 0) {
                i = ((i > this.buf.capacity()) ? this.buf.capacity() : i);
                this.buf.clear();
                this.buf.put(src, offset, i);
                i -= i;
                offset += i;
                this.buf.flip();
                if (SSLStreams.this.sendData(this.buf).result.getStatus() == SSLEngineResult.Status.CLOSED) {
                    this.closed = true;
                    if (i > 0) {
                        throw new IOException("output stream is closed");
                    }
                    continue;
                }
                else {
                    continue;
                }
            }
            return;
        }
        
        @Override
        public void flush() throws IOException {
        }
        
        @Override
        public void close() throws IOException {
            SSLStreams.WrapperResult wrapAndSend = null;
            SSLStreams.this.engine.closeOutbound();
            this.closed = true;
            SSLEngineResult.HandshakeStatus handshakeStatus = SSLEngineResult.HandshakeStatus.NEED_WRAP;
            this.buf.clear();
            while (handshakeStatus == SSLEngineResult.HandshakeStatus.NEED_WRAP) {
                wrapAndSend = SSLStreams.this.wrapper.wrapAndSend(this.buf);
                handshakeStatus = wrapAndSend.result.getHandshakeStatus();
            }
            assert wrapAndSend.result.getStatus() == SSLEngineResult.Status.CLOSED;
            return;
        }
    }
}
